#include <Arduino.h>
#include <Wire.h>
#include "INA219.h"

INA219::INA219(const uint8_t address, const uint8_t busVoltageRange, const uint16_t svr, const float shuntResistance) { // constructor
  i2cAddr = address;
  // shunt voltage is returned, regardless of range, in units of 0.01 mV.  Calculate what to multiply this by to convert to mA
  currentMultiplier = 0.01 / shuntResistance;
  // the config value sets the selected ranges for 12-bit resolution, 532us conversion time, shunt and bus continuous conversion
  // ina219config = svr == 40 ? 0x019F : svr == 80 ? 0x099F : svr == 160 ? 0x119F : 0x199F;
  // the config value sets the selected ranges for 128 samples.  conversion time 68.10 ms, shunt and bus continuous conversion
  ina219config = svr == 40 ? 0x07FF : svr == 80 ? 0x0FFF : svr == 160 ? 0x17FF : 0x1FFF;
  if (busVoltageRange != 16) ina219config |= 0x2000;
};

void INA219::start(void) {
    writeReg(0, ina219config);
    float f = 40960.0 * currentMultiplier;
    uint16_t i = (uint16_t)f;
    writeReg(5, i); 
}

float INA219::busVoltage_V(void) { // return voltage present at Vin-
  uint16_t bvr = readReg(2); // read bus voltage register
  float f = (float)(bvr &= 0xFFF8); // mask out 3 least significant bits (unused, conversion ready, math overflow)
  // least significant bit of converted value is always 4mV (but only accurate to 8mV on larger range).  Shifted left three bits, means that each count of 8 represents 4mV
  return f * 5.0E-4;
}

float INA219::current_mA(void) {
//  int16_t svr = (int16_t)(readReg(1)); // read shunt voltage register which is in 2's complement format
//  float f = (float)svr; // regardless of range selected, ina219 returns shunt voltage in units of 0.01 mV: the highest range goes up to +/-32000, the lowest +/-4000. Accuracy is only 1 part in 4000 of selected range.
//  return f * currentMultiplier;
  int16_t cur = (int16_t)(readReg(4)); // read current register which is in 2's complement format. Calibration register has been set to make this go in steps of 0.1 mA
  float f = (float)cur; 
  return f * 0.1;
}

void INA219::writeReg(uint8_t addr, uint16_t value) {
  Wire.beginTransmission(i2cAddr);
  Wire.write(addr);
  Wire.write((uint8_t)(value >> 8));
  Wire.write((uint8_t)(value & 0xFF));
  Wire.endTransmission();
  delay(1);
}

uint16_t INA219::readReg(uint8_t addr) {
  Wire.beginTransmission(i2cAddr);
  Wire.write(addr);
  Wire.endTransmission();
  Wire.requestFrom((int)i2cAddr, 2); // request two bytes from currently selected address
  uint16_t value = Wire.read();
  value <<= 8;
  value |= Wire.read();
  Wire.endTransmission();
  delay(1);
  return value;
}
