// simple single cell Lithium battery charger/discharger
// based on INA219DC I2C current sensor module
// I2C driven 20x4 LCD display
// uses nominal 5V supply to charge battery through dropper resistor
// discharge is via a second dropper resistor to ground
// uses separate relays in charge and discharge lines
// ceptimus August 2018

#define VERSION F("1.0.0")

// these voltages are Mike's choices.  Alter them to something different if you wish.
#define CHARGE_VOLTAGE 4.20
#define DISCHARGE_VOLTAGE 3.40
#define STORAGE_VOLTAGE 3.85

// 20x4 LCD I2C settings.  These will require changing if you use a different I2C adapter.  I used a red board with labels LCM1602 IIC HOYA Funduino, with a PCF8574 chip
#define LCD_CONFIGURE 0x27, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE

// INA219 settings.  These will require editing to give bigger current ranges, or to set an exact shunt resistance to get the most accurate current measurement
#define INA219_ADR 0x40

// shunt resistor fitted is an "R100" (0.1 ohm) but the exact value can be entered here - best way is to pass a steady current through INA219, measure with a trusted ammeter and tweak this value to get displayed current correct
#define INA219_SHUNT_RESISTANCE 0.1039

// only valid choices for INA219_BUS_VOLTAGE_RANGE are 16 (for 16V range with 4 mV accuracy) or 32 (for 32V range with 8 mV accuracy)
#define INA219_BUS_VOLTAGE_RANGE 16 
// the following shunt voltage ranges are possible, with the table indicating the nominal current range and resolution with the default R100 (0.1 ohm) shunt resistor
//  range (mV)  max current (mA)  resolution (mA)
//   +/-  40       +/-  400           0.1
//   +/-  80       +/-  800           0.2
//   +/- 160       +/- 1600           0.4
//   +/- 320       +/- 3200           0.8
// only valid choices for SHUNT_VOLTAGE_RANGE are 40, 80, 160, or 320.
#define INA219_SHUNT_VOLTAGE_RANGE 40

// pin definitions for I/O
#define CHARGE_RELAY 6
#define DISCHARGE_RELAY 7
#define BUTTON_1 3
#define BUTTON_2 4

// power supply input voltage into A0 via potential divider 
// V.psu   ----/\/\/\/\----+----/\/\/\/\---- GND
//                R1       |       R2 
//                        A0
// Recommended values:
//                     R1 = 68K
//                     R2 = 10K
// Gives theoretical input range of 1.1 * (68 + 10) / 10 = 8.58V  good for up to a 2-cell LiPo.  Adjust R1 and/or R2 if different power supply used.  Tweak values if desired to get accurate agreement with trusted DVM
// Note: a 2-cell LiPo should NOT be used if, like me, you used a 3.3V Arduino, and therefore chose to drive the relay coils straight from V.psu (into JD-VCC pin on relay board).  Standard SRD-05VDC-SL-C relays have coils rated 3.75V to 6V
// If you use a 5V Arduino then you can run the relay coils from the Arduino's VCC (jumper fitted on relay board to link VCC with JD_VCC).
// Choice of Arduino depends on the power source you intend to use.  If you have a phone charger or powerbank that delivers a steady 5V, then a 5V Arduino is a good idea, but if you want to use a 4-cell NiMH battery then a
// 3.3V one is better because it will still be running in specification, and able to power the LCD backlight with a regulated 3.3V, as the NiMH battery runs down.

#define PSU_VOLTAGE A0
#define R1 68000.0
#define R2 10000.0

#define SMOOTHED_ADC_COUNT_TO_VOLTS (1.1 * (R1 + R2) / R2 / 32736.0)

// debounce milliseconds for buttons (maximum 255)
#define DEBOUNCE_MS 150

// relay on/off state.  My relay board needed the Arduino pins high to leave the relays in the same state they are with no power present
#define RELAY_OFF HIGH
#define RELAY_ON LOW

#include <Wire.h> // standard Arduino I2C library
#include <LiquidCrystal_I2C.h> // Newliquidcrystal_1.3.5 folder (version at time of writing) should be in your (documents) Arduino/libraries folder.
// Download from https://bitbucket.org/fmalpartida/new-liquidcrystal/downloads/ or http://mode-zero.uk/viewtopic.php?f=24&t=417

#include "INA219.h" // custom INA219 class included with this sketch

LiquidCrystal_I2C lcd(LCD_CONFIGURE);
INA219 ina219(INA219_ADR, INA219_BUS_VOLTAGE_RANGE, INA219_SHUNT_VOLTAGE_RANGE, INA219_SHUNT_RESISTANCE);

uint32_t prev_millis;
uint16_t ms, minutes;
uint8_t seconds;
bool clockRunning; // true to make clock run
bool clockShowing; // true to make clock display (top right of display)
uint8_t mode = 0;
bool VAmAhShowing; // true to make 2nd row of display show either "  Connect battery   " when INA219 Vin- <= 1.0 V or "x.xx V        x.x mA" when V > 1.0 V.  Also show mAh and PSU voltage on 3rd row

uint16_t msTimer;

#define REFRESH_MILLIS 500
int16_t refreshVoltsAmpsTmr; // prevent display flicker by only refreshing display of INA219 voltage and current every REFRESH_MILLIS interval
bool toggle; // used to flash display elements for warnings or to gain attention

float targetVoltage;
bool charging; // true if charging, false if discharging

float mAs; // milliamp seconds - used for greater accumulating accuracy.  When abs(mAs) >= 3600.0, mAh is incremented/decremented and 3600.0 mAs are subtracted/added
int16_t mAh; // milliamp hours

uint16_t filteredPSU; // remove noise from analog input by simple 1st order low pass filter ("leaky integrator")

void setup() {
  digitalWrite(CHARGE_RELAY, RELAY_OFF);
  pinMode(CHARGE_RELAY, OUTPUT);
  digitalWrite(DISCHARGE_RELAY, RELAY_OFF);
  pinMode(DISCHARGE_RELAY, OUTPUT);
  pinMode(BUTTON_1, INPUT_PULLUP);
  pinMode(BUTTON_2, INPUT_PULLUP);

  analogReference(INTERNAL); // use the Arduino's built-in analog reference voltage of 1.1V - so the measuring range of an analog input will now be 0 to 1.1V
  delay(100); // allow analog input time to settle
  filteredPSU = analogRead(PSU_VOLTAGE) << 5;  // initialize filter
  
  Wire.begin();
  Wire.setClock(400000);
 
  lcd.begin(20, 4);
  ina219.start();
  prev_millis = millis();
  ms = minutes = 0;
  seconds = 0;
  clockRunning = clockShowing = false;
  mode = 0; // start up message
}

void loop() { 
  // run/display clock
  uint32_t now = millis();
  uint8_t loop_ms = (uint8_t)(now - prev_millis);
  ms += loop_ms;
  prev_millis = now;
  if (clockRunning) {
    if (ms >= 1000) {
      ms -= 1000;
      if (++seconds > 59) {
        seconds = 0;
        ++minutes;
      }
    }
  }
  else { // clock not running
    ms = 0;
  }

  uint8_t button = checkButtons(loop_ms);

  if (clockShowing) {
    lcd.setCursor(13, 0);
    if (minutes < 1000) lcd.print(" ");
    if (minutes < 100) lcd.print(" ");
    if (minutes < 10) lcd.print(" ");
    lcd.print(minutes);
    lcd.print(":");
    if (seconds < 10) lcd.print("0");
    lcd.print(seconds);
  }

  float busVoltage_V = ina219.busVoltage_V();
  float current_mA = ina219.current_mA();

  // update analog input filtered value - this is bit-shifted 5 bits so have range 0 - 1023 * 2^5 = 0 - 32736
  filteredPSU -= filteredPSU >> 5;
  filteredPSU += analogRead(PSU_VOLTAGE);
  float psuVoltage = filteredPSU * SMOOTHED_ADC_COUNT_TO_VOLTS;
  
  if (VAmAhShowing) {
    if (refreshVoltsAmpsTmr <= 0) {
      lcd.setCursor(0, 1);
      if (busVoltage_V <= 1.0) {
        lcd.print(F("  Connect battery   "));
      } else {
        lcd.print(busVoltage_V, 2);
        lcd.print(F(" V    "));
        lcd.setCursor(10, 1);
        if (abs(current_mA) < 999.95) {
          lcd.print(" ");
          if (abs(current_mA) < 99.95) {
            lcd.print(" ");
            if (abs(current_mA) < 9.95) {
              lcd.print(" ");
            }
          }
        }
        lcd.print(current_mA < 0.0 ? "-" : " ");
        lcd.print(abs(current_mA), 1);
        lcd.print(" mA");
      }
      float mAh_f = mAh;
      mAh_f += mAs / 3600.0;
      lcd.setCursor(0, 2);
      if (abs(mAh_f) < 999.95) {
        lcd.print(" ");
        if (abs(mAh_f) < 99.95) {
          lcd.print(" ");
          if (abs(mAh_f) < 9.95) {
            lcd.print(" ");
          }
        }
      }
      lcd.print(mAh_f < 0.0 ? "-" : " ");
      lcd.print(abs(mAh_f), 1);
      lcd.print(F(" mAh "));
      lcd.setCursor(14, 2);
      if (((psuVoltage < 4.2) || (psuVoltage > 6.0)) && toggle) { // flash psu voltage display if too low to charge battery or higher than the relay coil rating
        lcd.print(F("      "));
      } else {
        lcd.print(psuVoltage, 2);
        lcd.print(" V");
      }        
      
      refreshVoltsAmpsTmr += REFRESH_MILLIS;
      toggle = !toggle;
      mAs += ((float)REFRESH_MILLIS) / 1000.0 * current_mA;
      if (mAs >= 3600.0) {
        mAs -= 3600.0;
        ++mAh;
      } else if (mAs <= -3600.0) {
        mAs += 3600.0;
        --mAh;
      }
    }
    refreshVoltsAmpsTmr -= loop_ms;
  } else {
    refreshVoltsAmpsTmr = 0;
  }

  switch (mode) {
    case 0: // power on
      lcd.clear();
      clockRunning = clockShowing = VAmAhShowing = false;
      ms = minutes = 0;
      seconds = 0;
      lcd.print(F(" 1-Cell Li charger"));
      lcd.setCursor(0, 1);
      lcd.print(F("ceptimus.  Aug. 2018"));
      lcd.setCursor(3, 2);
      lcd.print(F("Version: "));
      lcd.print(VERSION);
      msTimer = 3500; // display power-on message for 3.5 secs
      ++mode;
      break;
    case 10: // clear display reset clock to 0:00, switch on 2nd line of display
      lcd.clear();
      VAmAhShowing = true;
      ms = minutes = 0;
      seconds = 0;
      ++mode;
      break;
    case 20: // select operation (Charge, Discharge, Storage)
      lcd.setCursor(0, 0);
      lcd.print(F("Operation: Charge   "));
      labelButtons("Select", "Start");
      ++mode;
      break;
    case 21: // check buttons - begin charge or swap to selecting of Discharge mode
      if (button == 1) {
        ++mode;  
      } else if (button == 2) {
        mode = 30;
      }
      break;
    case 22:
      lcd.setCursor(11, 0);
      lcd.print(F("Discharge"));
      ++mode;
      break;
    case 23: // check buttons - begin discharge or swap to selecting of Storage mode
      if (button == 1) {
        ++mode;  
      } else if (button == 2) {
        mode = 50;
      }
      break;
    case 24:
      lcd.setCursor(11, 0);
      lcd.print(F("Storage  "));
      ++mode;
      break;
    case 25: // check buttons - begin storage or swap to selecting of Charge mode
      if (button == 1) {
        mode = 20;  
      } else if (button == 2) {
        mode = 40;
      }
      break;
    case 30: // start charging
      lcd.clear();
      lcd.print(F("Charging"));
      targetVoltage = CHARGE_VOLTAGE;
      charging = true;
      mode = 60;
      break;
    case 40: // start storage charging/discharging
      lcd.clear();
      lcd.print(F("Storage"));
      targetVoltage = STORAGE_VOLTAGE;
      charging = busVoltage_V < targetVoltage;
      mode = 60;
      break;
    case 50: // start discharging
      lcd.clear();
      lcd.print(F("Discharging"));
      targetVoltage = DISCHARGE_VOLTAGE;
      charging = false;
      mode = 60;
      break;
    case 60: // reset mAh accumulator - begin charge/discharge process
      labelButtons("Stop", "");
      clockRunning = clockShowing = true;
      mAh = 0;
      mAs = 0.0;
      if (charging) {
        digitalWrite(CHARGE_RELAY, RELAY_ON);
      } else {
        digitalWrite(DISCHARGE_RELAY, RELAY_ON);
      }
      ++mode;
      break;
    case 61:
      if (charging) {
        if (busVoltage_V > targetVoltage) {
          mode = 70;
        }
      } else { // discharging
        if (busVoltage_V < targetVoltage) {
          mode = 70;
        }
      }
      if (button == 1) { // stop button pressed
        mode = 70;
      }
      break;
    case 70: // end charge or discharge process
      digitalWrite(CHARGE_RELAY, RELAY_OFF);
      digitalWrite(DISCHARGE_RELAY, RELAY_OFF);
      lcd.setCursor(0, 0);
      lcd.print(F("Complete   "));
      clockRunning = false;
      labelButtons("Reset", "Reset");
      ++mode;
      break;
    case 71: // wait for reset button
      lcd.setBacklight(toggle);
      if (button) {
        lcd.setBacklight(HIGH);
        clockShowing = false;
        mode = 10;
      }
      break;   
    default: // run msTimer. when elapsed advance to next mode decade.
      if (msTimer) {
        if (loop_ms > msTimer) {
          msTimer = 0;
        } else {
          msTimer -= loop_ms;
        }
      } else { // elapsed
        mode += 9;
        mode -= mode % 10;
      }
      break;
  }
}

void labelButtons(char *left, char *right) {
  lcd.setCursor(0, 3);
  lcd.print(F("                    "));
  lcd.setCursor(5 - ((strlen(left) + 1) >> 1), 3);
  lcd.print(left);
  lcd.setCursor(15 - ((strlen(right) + 1) >> 1), 3);
  lcd.print(right);
}

uint8_t checkButtons(uint8_t loop_ms) { // returns debounced button state: 0 (neither), 1 (left), 2 (right).  ms parameter is elapsed milliseconds since previous call
  static uint8_t buttonOffTmr[2] = { DEBOUNCE_MS, DEBOUNCE_MS };
  uint8_t button = 0;

  if (digitalRead(BUTTON_1) == LOW) { // pressed
    if (!buttonOffTmr[0]) { // debounce time complete
      button = 1;      
    }
    buttonOffTmr[0] = DEBOUNCE_MS;
  } else { // released - run debounce timer
    if (buttonOffTmr[0]) {
      if (loop_ms >= buttonOffTmr[0]) {
        buttonOffTmr[0] = 0;
      } else {
        buttonOffTmr[0] -= loop_ms;
      }
    }
  }

  if (digitalRead(BUTTON_2) == LOW) { // pressed
    if (!buttonOffTmr[1]) { // debounce time complete
      button = 2;      
    }
    buttonOffTmr[1] = DEBOUNCE_MS;
  } else { // released - run debounce timer
    if (buttonOffTmr[1]) {
      if (loop_ms >= buttonOffTmr[1]) {
        buttonOffTmr[1] = 0;
      } else {
        buttonOffTmr[1] -= loop_ms;
      }
    }
  }
  
  return button;
}

